//*****************************************************************************
//	
//	Copyright 2011 by WinSystems Inc.
//
//	Permission is hereby granted to the purchaser of WinSystems GPIO cards 
//	and CPU products incorporating a GPIO device, to distribute any binary 
//	file or files compiled using this source code directly or in any work 
//	derived by the user from this file. In no case may the source code, 
//	original or derived from this file, be distributed to any third party 
//	except by explicit permission of WinSystems. This file is distributed 
//	on an "As-is" basis and no warranty as to performance or fitness of pur-
//	poses is expressed or implied. In no case shall WinSystems be liable for 
//	any direct or indirect loss or damage, real or consequential resulting 
//	from the usage of this source code. It is the user's sole responsibility 
//	to determine fitness for any considered purpose.
//
//*****************************************************************************
//
//	Name	 : pcmvdxpwm.c
//
//	Project	 : PCM-VDX PWM Linux Driver
//
//	Author	 : Paul DeMetrotion
//
//*****************************************************************************
//
//	  Date		Revision	                Description
//	--------	--------	---------------------------------------------
//	05/12/11	  1.00		Original Release
//
//*****************************************************************************

static char *RCSInfo = "$Id: pcmvdxpwm.c,v 1.0 2011-05-11 paul Exp $";

#include <linux/autoconf.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/version.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/cdev.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include "pcmvdxpwm.h"	// our own ioctl numbers

#define DRV_NAME        "pcmvdxpwm"
#define DRV_VERSION     "1.0"
#define DRV_RELDATE     "12May2011"

// migration helpers taken from linux kernel source for 2.6.19 
// so out of tree code will build without errors or warnings
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19))
	#define IRQF_DISABLED			SA_INTERRUPT
	#define IRQF_SAMPLE_RANDOM		SA_SAMPLE_RANDOM
	#define IRQF_SHARED				SA_SHIRQ
	#define IRQF_PROBE_SHARED		SA_PROBEIRQ
	#define IRQF_PERCPU				SA_PERCPU

	#define IRQF_TRIGGER_LOW		SA_TRIGGER_LOW
	#define IRQF_TRIGGER_HIGH		SA_TRIGGER_HIGH
	#define IRQF_TRIGGER_FALLING	SA_TRIGGER_FALLING
	#define IRQF_TRIGGER_RISING		SA_TRIGGER_RISING
	#define IRQF_TRIGGER_MASK		SA_TRIGGER_MASK
#endif

//#define DEBUG			0			// uncomment for debug mode
#define SUCCESS			0
#define BUF_LEN			20
#define MAX_INTS		1024		// maximum number of buffered events
#define	VDX_SB_VENID	0x17F3
#define	VDX_SB_DEVID	0x6031

// interrupt handler
// the structure of the interrupt handler call changed with the 2.6.19 kernel
// this allows builds to succeed for either case
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19))
   irqreturn_t handler_1(int, void *, struct pt_regs *);
#else
   irqreturn_t handler_1(int, void *);
#endif

// common interrupt handler
void common_handler(void);

// Driver major number
static int pcmvdxpwm_init_major = 0;	// 0 = allocate dynamically
static int pcmvdxpwm_major;

// pcmvdxpwm device structure
struct pcmvdxpwm_dev {
	struct cdev cdev;	// char device structure
} *pcmvdxpwm_device;

// device specific
unsigned int pwm_base_addr = 0;	// io base address
unsigned int_count = 0;

// Our insmod command line arguments
static short ioaddr = 0;
static short irq = 0;
static short pwmclk = 0;
module_param(ioaddr, short, S_IRUGO);
module_param(irq, short, S_IRUGO);
module_param(pwmclk, short, S_IRUGO);

// interrupt buffer
unsigned int_buffer[MAX_INTS];
int inptr = 0;		// queue pointers
int outptr = 0;		//		"

// create the wait queue
static DECLARE_WAIT_QUEUE_HEAD(wq);

/***************************************************************************
*			DEVICE_SUBROUTINES
***************************************************************************/

u32 read_sb(int idx)
{
	struct pci_dev *dev;
	u32 data;

	// find south bridge
	dev = pci_get_device(VDX_SB_VENID, VDX_SB_DEVID, NULL);

	// read config reg
	pci_read_config_dword(dev, idx, &data);

	#ifdef DEBUG
	printk("<1>PCMVDXPWM - SB Read Offset 0x%02X Data 0x%08X\n", idx, data);
	#endif
		
	// release pci device
	pci_dev_put(dev);

	return data;
}

void write_sb(int idx, u32 val)
{
	struct pci_dev *dev;

	// find south bridge
	dev = pci_get_device(VDX_SB_VENID, VDX_SB_DEVID, NULL);

	// write config reg
	pci_write_config_dword(dev, idx, val);
		
	#ifdef DEBUG
	printk("<1>PCMVDXPWM - SB Write Offset 0x%02X Data 0x%08X\n", idx, val);
	#endif
		
	// release pci device
	pci_dev_put(dev);
}

void pwm_clr_int(int channel)
{
	// write a one to selected channel to clear interrupt
	outw(1 << channel, pwm_base_addr + PWM_INT_STAT);
}

int pwm_get_int(void)
{
	int temp, x;

	// read port 0 interrupt status
	temp = inw(pwm_base_addr + PWM_INT_STAT);

	// see if any bit is set, if so return the channel
	if (temp !=0)
	{
		for(x=0; x <=15; x++)
		{
			if(temp & (1 << x))
			{
				return(x + 1);	// return channel number with active int
			}
		}
	}

	// no interrupts so
	return 0;
}

int pwm_get_buffered_int(void)
{
	int intr_chan;

	// If any pending, return the appropriate channel number
	if (irq == 0) 
	{
	    intr_chan = pwm_get_int();

		if (intr_chan)
			pwm_clr_int(intr_chan - 1);

		return intr_chan;
	}

	if (outptr != inptr)
	{
	    intr_chan = int_buffer[outptr++];
	
		if (outptr == MAX_INTS)
			outptr = 0;

		return intr_chan;
	}

	return 0;
}

/***************************************************************************
*			COMMON INTERRUPT HANDLER
***************************************************************************/

void common_handler(void)
{
	int intr_chan, count;

	while(1) 
	{
	    count = 0;

	    intr_chan = pwm_get_int();

		if (intr_chan)
		{
			pwm_clr_int(intr_chan - 1);

			#ifdef DEBUG
			printk("<1>PCMVDXPWM - Interrupt Channel %d\n", intr_chan - 1);
			#endif

			count++;
					
			int_buffer[inptr++] = intr_chan;

			if (inptr == MAX_INTS)
				inptr = 0;

			// Wake up anyone sleeping on this event
			wake_up_interruptible(&wq);
		}
		
		if (!count)
			break;
	}
}

/****************************************************************************
*			INTERRUPT HANDLER
****************************************************************************/

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19))

irqreturn_t handler_1(int pirq, void *dev_id, struct pt_regs *regs)
{
	#ifdef DEBUG
	printk("<1>PCMVDXPWM - Interrupt received\n");
	#endif

	common_handler();
	return(IRQ_HANDLED);
}

#else

irqreturn_t handler_1(int irq, void *dev_id)
{
	#ifdef DEBUG
	printk("<1>PCMVDXPWM - Interrupt received\n");
	#endif

	common_handler();
	return(IRQ_HANDLED);
}

#endif

/****************************************************************************
*			DEVICE OPEN
****************************************************************************/

static int device_open(struct inode *inode, struct file *file)
{
	#ifdef DEBUG
	printk("device_open(%p)\n", file);
	#endif

	return(SUCCESS);
}

/*************************************************************************
*			DEVICE CLOSE
**************************************************************************/

int device_release(struct inode *inode, struct file *file)
{
	#ifdef DEBUG
	printk("device_release(%p, %p)\n", inode, file);
	#endif 

	return(SUCCESS);
}

/****************************************************************************
*			DEVICE IOCTL
****************************************************************************/

int device_ioctl(struct inode *inode, struct file *file, unsigned int ioctl_num, unsigned long ioctl_param)
{
	unsigned long *wform;
	int channel, intr, clk;
	unsigned int addr, temp;
	unsigned long pci_dat;

	#ifdef DEBUG
	//printk("<1>PCMVDXPWM - IOCTL Call - Code %04X\n", ioctl_num);
	#endif

	// Switch according to the ioctl called
	switch (ioctl_num) 
	{
		case IOCTL_ENAB_PWM:
			wform = (unsigned long *) ioctl_param;

			// set base address for specific channel
			addr = pwm_base_addr + (0xC * (wform[0] + 1));

			// set pulse low count
			outl(wform[1], addr);
	
			// set pulse high count
			outl(wform[2], addr + 4);

			// enable channel
			outl(PWM_ENBL | PWM_CONT_MODE, addr + 8);	// continuous mode
			
			#ifdef DEBUG
			printk("<1>PCMVDXPWM - ioctl_enab_pwm Channel %ld Low PC %ld High PC %ld\n", \
				wform[0], wform[1], wform[2]);
			#endif

			return(SUCCESS);

		case IOCTL_ENAB_PWM_RC:
			wform = (unsigned long *) ioctl_param;

			// set base address for specific channel
			addr = pwm_base_addr + (0xC * (wform[0] + 1));

			// set pulse low count
			outl(wform[1], addr);
	
			// set pulse high count
			outl(wform[2], addr + 4);

			// enable channel
			outl(PWM_ENBL | wform[3], addr + 8);	// repeat count mode
			
			#ifdef DEBUG
			printk("<1>PCMVDXPWM - ioctl_enab_pwm_rc Channel %ld Low PC %ld High PC %ld Repeat %ld\n", \
				wform[0], wform[1], wform[2], wform[3]);
			#endif

			return(SUCCESS);

		case IOCTL_DISAB_PWM:
			channel = ioctl_param & 0xff;

			// set base address for specific channel
			addr = pwm_base_addr + (0xC * (channel + 1));

			// set pulse low count
			outl(0, addr);
	
			// set pulse high count
			outl(0, addr + 4);

			// enable channel
			outl(0, addr + 8);	// continuous mode
			
			#ifdef DEBUG
			printk("<1>PCMVDXPWM - ioctl_disab_pwm Channel %d\n", channel);
			#endif

			return(SUCCESS);

		case IOCTL_ENAB_PWM_INT:
			channel = ioctl_param & 0xff;

			// retrieve interrupt mask reg
			temp = inw(pwm_base_addr + PWM_INT_MSK);

			// set the enable bit for our channel number
			temp |= (1 << channel);

			// write new value
			outw(temp, pwm_base_addr + PWM_INT_MSK);

			#ifdef DEBUG
			printk("<1>PCMVDXPWM - ioctl_enab_pwm_int Channel %d\n", channel);
			#endif

			return(SUCCESS);

		case IOCTL_DISAB_PWM_INT:
			channel = ioctl_param & 0xff;

			// retrieve interrupt mask reg
			temp = inw(pwm_base_addr + PWM_INT_MSK);

			// clear the enable bit for our channel number
			temp &= ~(1L << channel);

			// write new value
			outw(temp, pwm_base_addr + PWM_INT_MSK);

			#ifdef DEBUG
			printk("<1>PCMVDXPWM - ioctl_disab_pwm_int Channel %d\n", channel);
			#endif

			return(SUCCESS);

		case IOCTL_CLR_PWM_INT:
			pwm_clr_int(ioctl_param & 0xff);

			#ifdef DEBUG
			printk("<1>PCMVDXPWM - ioctl_clr_pwm_int Channel %d\n", (int) ioctl_param & 0xff);
			printk("<1>PCMVDXGPIO - int_count = %d\n", int_count);
			#endif

			return(SUCCESS);

		case IOCTL_GET_PWM_INT:
			intr = pwm_get_buffered_int();

			#ifdef DEBUG
			printk("<1>PCMVDXPWM - ioctl_get_pwm_int Status %d\n", intr);
			#endif
	
			return(intr);

	    case IOCTL_WAIT_PWM_INT:
			#ifdef DEBUG
			printk("<1>PCMVDXPWM - ioctl_wait_pwm_int\n");
			#endif

			// Return one if there is one there
			intr = pwm_get_buffered_int();

			if (intr)
				return(intr);

			// None pending, so sleep awaiting an interrupt
			#ifdef DEBUG
			printk("<1> PCMVDXPWM : current process %i (%s) going to sleep\n", current->pid, current->comm);
			#endif

			interruptible_sleep_on(&wq);

			// Process awakens and resume execution here

			#ifdef DEBUG
			printk("<1> PCMVDXPWM : awoken %i (%s)\n", current->pid, current->comm);
			#endif

			// Getting here does not guarantee that there's an in interrupt available
			// we may have been awakened by some other signal. In any case We'll 
			// return whatever's available in the interrupt queue even if it's empty
			intr = pwm_get_buffered_int();

			return(intr);

		case IOCTL_PWM_CLK_SEL:
			clk = ioctl_param & 0xff;

			pci_dat = read_sb(PWM_CTRL);

			if (clk)
				write_sb(PWM_CTRL, pci_dat | (1 << 24));
			else
				write_sb(PWM_CTRL, pci_dat & ~(1 << 24));

			#ifdef DEBUG
			printk("<1>PCMVDXPWM - ioctl_pwm_clk_sel Clk %d\n", clk);
			#endif

			return(SUCCESS);

		// Catch all return
		default:
			return(-EINVAL);
	}
}

/****************************************************************************
*			MODULE DECLARATIONS
*****************************************************************************/
// This structure will hold the functions to be called 
// when a process does something to our device

struct file_operations pcmvdxpwm_fops = {
	owner:		THIS_MODULE,
	llseek:		NULL,
	read:		NULL, 
	write:		NULL,
	ioctl:		device_ioctl,
	open:		device_open,
	release:	device_release,
};

/****************************************************************************
*			INIT_MODULE
*****************************************************************************/

int init_module()
{
	struct pcmvdxpwm_dev *dev;
	dev_t devno;
	int ret_val;
	unsigned long byMask;

	// Let the world know we are here
	printk("<1>WinSystems, Inc PCMVDXPWM PWM Linux Device Driver\n");
	printk("<1>Copyright 2011, All Rights Reserved\n");
	printk("<1>%s\n",RCSInfo);

	// allocate memory for the device structure
	pcmvdxpwm_device = kmalloc(sizeof(struct pcmvdxpwm_dev), GFP_KERNEL);

	if (pcmvdxpwm_device == NULL)
	{
		printk("<1>PCMVDXPWM - No memory resources for device structure\n");
		return -ENOMEM;
	}

	// Register the character device
	if(pcmvdxpwm_init_major)
	{
		pcmvdxpwm_major = pcmvdxpwm_init_major;
		devno = MKDEV(pcmvdxpwm_major, 0);
		ret_val = register_chrdev_region(devno, 1, DRV_NAME);
	}
	else
	{
		ret_val = alloc_chrdev_region(&devno, 0, 1, DRV_NAME);
		pcmvdxpwm_major = MAJOR(devno);
	}

	// Negative values signify an error
	if (ret_val < 0) 
	{
		printk("<1>PCMVDXPWM - Sorry, registering the character device failed with %d\n", ret_val);
		goto exit_free;
	}
	else
		printk("<1>PCMVDXPWM - Major number %d assigned\n", pcmvdxpwm_major);

	dev = pcmvdxpwm_device;		// only one device for now

	// initialize character device
	cdev_init(&dev->cdev, &pcmvdxpwm_fops);
	dev->cdev.owner = THIS_MODULE;
	dev->cdev.ops = &pcmvdxpwm_fops;
	ret_val = cdev_add(&dev->cdev, devno, 1);

	if(ret_val)
	{
		printk("<1>PCMVDXPWM - Error %d adding %s character device\n", ret_val, DRV_NAME);
		goto exit_cdev_unregister;
	}
	else
		printk("<1>PCMVDXPWM - Added character device %s\n", DRV_NAME);


	// Check and Map our I/O region requests
	pwm_base_addr = ioaddr & 0xFE00;	// set base address

	if (request_region(pwm_base_addr, 0x190, "PCMVDXPWM") == NULL)
	{
		printk("<1>PCMVDXPWM - Unable to use I/O Address %04X\n", pwm_base_addr);
		goto exit_cdev_delete;
	}
	else
		printk("<1>PCMVDXPWM - Address %04X-%04Xh registered to device\n", pwm_base_addr, pwm_base_addr + 0x190);

	// If an interrupt is specified, install a handler for it
    if (irq != 0) 
	{
		if (request_irq(irq, handler_1, IRQF_SHARED, "PCMVDXPWM", RCSInfo))
		{
			printk("<1>PCMVDXPWM - Unable to register IRQ %d\n", irq);
			goto exit_release_io;
		}
		else
			printk("<1>PCMVDXPWM - IRQ %d registered to device\n", irq);
    }	

	// disable all interrupts
	outl(0, pwm_base_addr + PWM_INT_MSK);

	// clear all pending interrupts
	outl(0xFFFFFFFF, pwm_base_addr + PWM_INT_STAT);

	// set irq number according to argument
	switch(irq)
	{
		case 0: byMask = 0x00; break;
		case 9: byMask = 0x01; break;
		case 3: byMask = 0x02; break;
		case 10: byMask = 0x03; break;
		case 4: byMask = 0x04; break;
		case 5: byMask = 0x05; break;
		case 7: byMask = 0x06; break;
		case 6: byMask = 0x07; break;
		case 1: byMask = 0x08; break;
		case 11: byMask = 0x09; break;
		case 12: byMask = 0x0b; break;
		case 14: byMask = 0x0d; break;
		case 15: byMask = 0x0f; break;
		default: return 1;
	}

	write_sb(PWM_CTRL, (pwmclk << 24) | (1 << 23) | (byMask << 16) | pwm_base_addr);	// bit 23 = enable

	return(SUCCESS);

exit_release_io:
	release_region(pwm_base_addr, 0x190);
exit_cdev_delete:
	cdev_del(&dev->cdev);
exit_cdev_unregister:
	unregister_chrdev_region(devno, 1);
	pcmvdxpwm_major = 0;
exit_free:
	kfree(pcmvdxpwm_device);
	return -ENODEV;
}

/***************************************************************************
*			CLEANUP_MODULE
***************************************************************************/

void cleanup_module()
{
	// release irq
	if (irq)
		free_irq(irq, RCSInfo);

	// release I/O
	release_region(pwm_base_addr, 0x190);

	// remove and unregister the device
	cdev_del(&pcmvdxpwm_device->cdev);
	unregister_chrdev_region(MKDEV(pcmvdxpwm_major, 0), 1);
	pcmvdxpwm_major = 0;

	// release memory
	kfree(pcmvdxpwm_device);
}  

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Paul DeMetrotion");
MODULE_DESCRIPTION("WinSystems,Inc. PCM-VDX PWM Driver");

