//*****************************************************************************
//	
//	Copyright 2011 by WinSystems Inc.
//
//	Permission is hereby granted to the purchaser of WinSystems GPIO cards 
//	and CPU products incorporating a GPIO device, to distribute any binary 
//	file or files compiled using this source code directly or in any work 
//	derived by the user from this file. In no case may the source code, 
//	original or derived from this file, be distributed to any third party 
//	except by explicit permission of WinSystems. This file is distributed 
//	on an "As-is" basis and no warranty as to performance or fitness of pur-
//	poses is expressed or implied. In no case shall WinSystems be liable for 
//	any direct or indirect loss or damage, real or consequential resulting 
//	from the usage of this source code. It is the user's sole responsibility 
//	to determine fitness for any considered purpose.
//
//*****************************************************************************
//
//	Name	 : pcmvdxpwmio.c
//
//	Project	 : PCM-VDX PWM Linux Driver
//
//	Author	 : Paul DeMetrotion
//
//*****************************************************************************
//
//	  Date		Revision	                Description
//	--------	--------	---------------------------------------------
//	05/12/11	  1.00		Original Release
//
//*****************************************************************************

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>      /* open */ 
#include <unistd.h>     /* exit */
#include <sys/ioctl.h>  /* ioctl */

// Include the WinSystems PCMVDXPWM definitions
#include "pcmvdxpwm.h"

// device handle
int check_handle(void);
int handle = 0;

// the names of our device nodes
char *device_id = "/dev/pcmvdxpwm";

//------------------------------------------------------------------------
//
// ioctl_enab_pwm - Enable continuous PWM on a single channel
//
// Arguments:
//			channel		PWM channel to enable (0-15)
//			low			Low pulse count
//			high		High pulse count 
//
// Returns:
//			SUCCESS	Completion
//			-1		Device handle is invalid
//
//------------------------------------------------------------------------
int ioctl_enab_pwm(int channel, unsigned long low, unsigned long high)
{
	if(check_handle())		// Check if device available
		return(-1);			// Return -1 if not

	wform[0] = (unsigned long) channel;
	wform[1] = low;
	wform[2] = high;

	return(ioctl(handle, IOCTL_ENAB_PWM, wform));
}

//------------------------------------------------------------------------
//
// ioctl_enab_pwm_rc - Enable single count PWM on a single channel
//
// Arguments:
//			channel		PWM channel to enable (0-15)
//			low			Low pulse count
//			high		High pulse count 
//			repeat		Number of pulses
//
// Returns:
//			SUCCESS	Completion
//			-1		Device handle is invalid
//
//------------------------------------------------------------------------
int ioctl_enab_pwm_rc(int channel, unsigned long low, unsigned long high, unsigned long repeat)
{
	if(check_handle())		// Check if device available
		return(-1);			// Return -1 if not

	wform[0] = (unsigned long) channel;
	wform[1] = low;
	wform[2] = high;
	wform[3] = repeat;

	return(ioctl(handle, IOCTL_ENAB_PWM_RC, wform));
}

//------------------------------------------------------------------------
//
// ioctl_disab_pwm - Disable PWM for a single channel
//
// Arguments:
//			channel		PWM channel to disable (0-15)
//
// Returns:
//			SUCCESS	Completion
//			-1		Device handle is invalid
//
//------------------------------------------------------------------------
int ioctl_disab_pwm(int channel)
{
	if(check_handle())		// Check if device available
		return(-1);			// Return -1 if not

	return(ioctl(handle, IOCTL_DISAB_PWM, channel));
}

//------------------------------------------------------------------------
//
// ioctl_enab_pwm_int - Enable PWM interrupt for a single channel
//
// Arguments:
//			channel		PWM channel to enable (0-15)
//
// Returns:
//			SUCCESS	Completion
//			-1		Device handle is invalid
//
//------------------------------------------------------------------------
int ioctl_enab_pwm_int(int channel)
{
	if(check_handle())		// Check if device available
		return(-1);			// Return -1 if not

	return(ioctl(handle, IOCTL_ENAB_PWM_INT, channel));
}

//------------------------------------------------------------------------
//
// ioctl_disab_pwm_int - Disable PWM interrupt for a single channel
//
// Arguments:
//			channel		PWM channel to disable (0-15)
//
// Returns:
//			SUCCESS	Completion
//			-1		Device handle is invalid
//
//------------------------------------------------------------------------
int ioctl_disab_pwm_int(int channel)
{
	if(check_handle())		// Check if device available
		return(-1);			// Return -1 if not

	return(ioctl(handle, IOCTL_DISAB_PWM_INT, channel));
}

//------------------------------------------------------------------------
//
// ioctl_clr_pwm_int - Clear PWM interrupt for a single channel
//
// Arguments:
//			channel		PWM channel to clear (0-15)
//
// Returns:
//			SUCCESS	Completion
//			-1		Device handle is invalid
//
//------------------------------------------------------------------------
int ioctl_clr_pwm_int(int channel)
{
	if(check_handle())		// Check if device available
		return(-1);			// Return -1 if not

	return(ioctl(handle, IOCTL_CLR_PWM_INT, channel));
}

//------------------------------------------------------------------------
//
// ioctl_get_pwm_int - Poll for notification of a PWM channel
//			state change.
//
// Arguments:
//
// Returns:
//			0		No interrupt
//			1-16	Channel of received interrupt
//						1-8 = Port0[0..7]
//						9-16 = Port1[0..7]
//			-1		Device handle is invalid
//
//------------------------------------------------------------------------
int ioctl_get_pwm_int(void)
{
	if(check_handle())		// Check if device available
		return(-1);			// Return -1 if not

	return(ioctl(handle, IOCTL_GET_PWM_INT, 0));
}

//------------------------------------------------------------------------
//
// ioctl_wait_pwm_int - Wait for the notification of an input point 
//			state change.
//
// Arguments:
//
//			0		No interrupt
//			1-16	Channel of received interrupt
//						1-8 = Port0[0..7]
//						9-16 = Port1[0..7]
//			-1		Device handle is invalid
//
//------------------------------------------------------------------------
int ioctl_wait_pwm_int(void)
{
	if(check_handle())		// Check if device available
		return(-1);			// Return -1 if not

	return(ioctl(handle, IOCTL_WAIT_PWM_INT, 0));
}

//------------------------------------------------------------------------
//
// ioctl_pwm_clk_sel - Toggle the PWM clock source
//
// Arguments:
//			clk		PWM clock 
//						0 = 10 MHz
//						1 = 50 MHz
//
// Returns:
//			SUCCESS	Completion
//			-1		Device handle is invalid
//
//------------------------------------------------------------------------
int ioctl_pwm_clk_sel(int clk)
{
	if(check_handle())		// Check if device available
		return(-1);			// Return -1 if not

	return(ioctl(handle, IOCTL_PWM_CLK_SEL, clk));
}

//------------------------------------------------------------------------
//
// check_handle - Checks that a handle to the device file exists.
//			If it does not a file open is performed.
//
// Arguments:
//
// Returns:
//			0		Device handle is valid
//			-1		Device handle is invalid
//
//------------------------------------------------------------------------

int check_handle(void)
{
    if(handle > 0)		// If it's already a valid handle
		return(0);		// return that fact

	if(handle == -1)	// If it's already been tried
		return(-1);		// return that fact

	// Perform an open on the device file and store the result
	// open returns -1 upon failure, a file descriptor (a small non-negative integer) on success
	handle = open(device_id, O_RDWR);

    if(handle)		// If it's now a valid handle
		return(0);	// return 0

    handle = -1;	// Flag that it is not valid

	return(-1);		// and return -1 to indicate that fact
}
