//*****************************************************************************
//	
//	Copyright 2011 by WinSystems Inc.
//
//	Permission is hereby granted to the purchaser of WinSystems GPIO cards 
//	and CPU products incorporating a GPIO device, to distribute any binary 
//	file or files compiled using this source code directly or in any work 
//	derived by the user from this file. In no case may the source code, 
//	original or derived from this file, be distributed to any third party 
//	except by explicit permission of WinSystems. This file is distributed 
//	on an "As-is" basis and no warranty as to performance or fitness of pur-
//	poses is expressed or implied. In no case shall WinSystems be liable for 
//	any direct or indirect loss or damage, real or consequential resulting 
//	from the usage of this source code. It is the user's sole responsibility 
//	to determine fitness for any considered purpose.
//
//*****************************************************************************
//
//	Name	 : poll.c
//
//	Project	 : PCM-VDX PWM Sample Application Program
//
//	Author	 : Paul DeMetrotion
//
//*****************************************************************************
//
//	  Date		Revision	                Description
//	--------	--------	---------------------------------------------
//	05/12/11	  1.00		Original Release
//
//*****************************************************************************

// This program demonstrates one manner in which an unprivledged 
//   application running in user space can sychronize to hardware events 
//   handled by a device driver running in Kernel space.

#include <stdio.h>
#include <fcntl.h>      /* open */ 
#include <unistd.h>     /* exit */
#include <sys/ioctl.h>  /* ioctl */
#include <stdlib.h>
#include <pthread.h>

#include "pcmvdxpwm.h"

// These two functions will be sub-processes using the Posix threads 
// capability of Linux. These two threads will simulate a type of 
// interrupt service routine in that they will start up and then suspend 
// until an interrupt occurs and the driver awakens them.
void *thread_function(void *arg);

// Event count, counts the number of events we've handled
volatile int event_count;
volatile int exit_flag = 0;

unsigned int BitNo = 0;
char line[80];

int main(int argc, char *argv[])
{
	unsigned int pwm_base_addr = 0x1200;
	int channel;
	int i, intr, res;
	pthread_t a_thread;

	if(argc != 2)
	{
		printf("\nUsage: %s <channel> \n", argv[0]);
		return 1;
	}
	else
	{
		printf("Test to verify proper PCM-VDX PWM interrupt operation\n");
		channel = atoi(argv[1]);
	}

	// clear out any events that are queued up within the 
    // driver and clear any pending interrupts
    while (intr = ioctl_get_pwm_int())
	{
		printf("Clearing interrupt on channel %d\n", intr-1);
		ioctl_clr_pwm_int(intr);
    }

	// set pwm clock to 50 MHz
	ioctl_pwm_clk_sel(PWM_50MHZ);

	// enable interrupt for chosen channel
	ioctl_enab_pwm_int(channel);

    // Now the sub-threads will be started
    printf("Splitting off polling process\n");

    res = pthread_create(&a_thread, NULL, thread_function, NULL);
	
	if(res != 0) 
	{
		perror("Thread creation failed\n");
		exit(EXIT_FAILURE);
    }

    // Both threads are now running in the background. They'll execute up
    // to the point were there are no interrupts and suspend. We as their
    // parent continue on. The nice thing about POSIX threads is that we're 
    // all in the same data space the parent and the children so we can 
    // share data directly. In this program we share the event_count 
    // variable.

	// We'll continue on in this loop until we're terminated
    while(1) 
	{
		// Print Something so we know the foreground is alive
		printf("Press 'I' to generate interrupt, 'Q' to quit\n");

		// The foreground will now wait for an input from the console
		// We could actually go on and do anything we wanted to at this 
		// point

		fgets(line, 75, stdin);

		// if we hit 'Q' and <Enter>, the program terminates
		if(line[0] == 'q' || line[0] == 'Q')
		    break;
		// Here's the actual exit. If we hit 'Q' and Enter. The program terminates.
		else if(line[0] == 'i' || line[0] == 'I')
		{
			ioctl_enab_pwm_rc(channel, 0xA, 0xA, 10);	// 2.5 MHz 50% DC 10 pulses
			printf("Channel %d -> 10 pulses, 2.5 MHz Waveform, 50%% DC\n", channel);
		}
    }

    // This flag is a shared variable that the children can look at to
	// know we're finished and they can exit too. */
    exit_flag = 1;

    // Display our event count total
    printf("Event count = %05d\r", event_count);
    printf("\nCanceling thread\n");
    
    // If our children are not in a position to see the exit_flag, we
    // will use a more forceful technique to make sure they terminate with
    // us. If we leave them hanging and we try to re-run the program or
    // if another program wants to talk to the device they may be locked
    // out. This way everything cleans up much nicer.

	// disable channel interrupts
	ioctl_disab_pwm_int(channel);

	pthread_cancel(a_thread);
    fflush(NULL);
}

// This is the first of the sub-processes. For the purpose of this
// example, it does nothing but wait for an interrupt and then 
// reports that fact via the console. It also increments the shared 
// data variable event_count.

void *thread_function(void *arg)
{
	int c;

	printf("Thread exists!\n");

	while(1) 
	{
	    pthread_testcancel();

		if(exit_flag)
			break;

	    // This call will put THIS process to sleep until either an
	    // interrupt occurs or a terminating signal is sent by the 
	    // parent or the system.
	    c = ioctl_wait_pwm_int();

	    // We check to see if it was a real interrupt instead of a
	    // termination request.
	    if(c > 0) 
		{
		    printf("Event sense occured on channel %d\n", c-1);

			++event_count;
	    }
		else
			break;
	}
}
